% lowlev.spc - interface specifications for `lowlev' abstractions
%
% This file was derived from PCLU distribution (as of 11/18/93), and
% has been modified by Department of Mathematical and Computing
% Sciences, Tokyo Institute of Technology.
%
% Conditions for using, copying, modifying, and distributing this
% file are the same as those of PCLU.
%
% $Id: lowlev.spc,v 2.0 1996/06/10 09:02:29 ushijima Exp $


%  Copyright	Massachusetts Institute of Technology     1989
% Copyright Massachusetts Institute of Technology 1981, 1989
% Copyright Massachusetts Institute of Technology 1982, 1989
% Copyright Massachusetts Institute of Technology 1983, 1989
% Copyright Massachusetts Institute of Technology 1985, 1989


% The following table shows the file in PCLU distribution from which the
% interface specification for each abstraction was derived.
%
%	================================================
%	Abstraction name	File name in PCLU
%	------------------------------------------------
%	_adv			lib/_adv.spc
%	_anyize			lib/etc.spc
%	_byteptr		lib/_byteptr.spc
%	_byteptr8		lib/_byteptr8.spc
%	_bytevec		lib/_bytevec.spc
%	_chan			lib/_chan.spc
%	_change_signal		lib/etc.spc
%	_cleanup_		sys/clu/_cleanup_
%	_cvt			lib/etc.spc
%	_dtable_size		lib/etc.spc
%	_environ		lib/etc.spc
%	_environs		lib/etc.spc
%	_ernum			lib/etc.spc
%	_erstr			lib/etc.spc
%	_event			lib/_event.spc
%	_fixup_dir_part		sys/clu/_fixup_dir_part
%	_fixup_file_name	sys/clu/_fixup_file_name
%	_free_space		lib/etc.spc
%	_gc			lib/_gc.spc
%	_get_gid		lib/etc.spc
%	_get_pid		lib/etc.spc
%	_get_priority		lib/etc.spc
%	_get_termcap		lib/etc.spc
%	_get_uid		lib/etc.spc
%	_groups			lib/etc.spc
%	_heap_size		lib/etc.spc
%	_home_dir		lib/etc.spc
%	_host_id		lib/etc.spc
%	_ioctl			lib/etc.spc
%	_job			lib/_job.spc
%	_last_indexc		lib/etc.spc
%	_lstat			lib/etc.spc
%	_obj			lib/_obj.spc
%	_page_size		lib/etc.spc
%	_pause			lib/etc.spc
%	_read_link		lib/etc.spc
%	_resolve		lib/clu/_resolve.clu
%	_round_real		lib/etc.spc
%	_rusage			lib/etc.spc
%	_select			lib/etc.spc
%	_set_alarm		lib/etc.spc
%	_set_limit		lib/etc.spc
%	_set_priority		lib/etc.spc
%	_signal_type		lib/clu/_signal_type.clu
%	_skip_fill		sys/clu/_skip_fill
%	_split_name_part	lib/etc.spc
%	_stat			lib/etc.spc
%	_tagcell		lib/_tagcell.spc
%	_termcap		sys/clu/_termcap
%	_termcap_decode		sys/clu/_termcap
%	_termcap_entry		lib/etc.spc
%	_umask			lib/etc.spc
%	_unparse_real		lib/etc.spc
%	_vec			lib/_vec.spc
%	_wordvec		lib/_wordvec.spc
%	_working_dir		sys/clu/_working_dir
%	gcd_store		lib/gcd_tab.spc
%	gcd_tab			lib/gcd_tab.spc
%	inet_address		lib/clu/inet_address.clu
%	istream2chan		lib/etc.spc
%	stream2chan		lib/etc.spc
%	------------------------------------------------


% Data abstractions


% _Adv

_adv = cluster[T: type] is create, equal, get_low, set_low, get_size, set_size,
       get_start, set_start, get_predict, set_predict, get_vector, set_vector,
       _gcd

% Operations

create = proc(LOW: int, PRED: int, V: _vec[T]) returns(_adv[T]) signals(bounds)
	% effects  Signals `bounds' if LOW is greater than an implementation-
	%	   defined limit; otherwise returns a new object such that:
	%		1. low bound is LOW
	%		2. predicted size is PRED
	%		3. elements are those of V
	end create

equal = proc(ADV1, ADV2: _adv[T]) returns(bool)
	% effects  Returns `true' if ADV1 and ADV2 are the same object;
	%	   otherwise returns `false'.
	end equal

get_low = proc(ADV: _adv[T]) returns(int)
	% effects  Returns the low bound of ADV.
	end get_low

set_low = proc(ADV: _adv[T], LOW: int) signals(bounds)
	% modifies ADV
	% effects  Signals `bounds' if LOW is out of implementation-defined
	%	   bounds; otherwise sets the low bound of ADV to LOW.
	end set_low

get_size = proc(ADV: _adv[T]) returns(int)
	% effects  Returns the size (i.e., the number of the elements) of ADV.
	end get_size

set_size = proc(ADV: _adv[T], SIZE: int) signals(bounds)
	% modifies ADV
	% effects  ...
	end set_size

get_start = proc(ADV: _adv[T]) returns(int)
	% effects  Returns the index where the first element of ADV is stored
	%	   in the internal vector of ADV.
	end get_start

set_start = proc(ADV: _adv[T], STRT: int) signals(bounds)
	% modifies ADV
	% effects  set index in real part of first element(recomputes size)
	end set_start

get_predict = proc(ADV: _adv[T]) returns(int)
	% effects  Returns the predicted size of ADV.
	end get_predict

set_predict = proc(ADV: _adv[T], PRED: int)
	% modifies ADV
	% effects  Sets the predicted size of ADV.
	end set_predict

get_vector = proc(ADV: _adv[T]) returns(_vec[T])
	% effects  Returns the internal vector of ADV.
	end get_vector

set_vector = proc(ADV: _adv[T], V: _vec[T]) signals(bounds)
	% modifies ADV
	% effects  Changes ADV so that the internal vector of ADV becomes V.
	end set_vector

_gcd = proc(ADV: _adv[T], TAB: gcd_tab) returns(int)
       where T has _gcd: proctype(T, gcd_tab) returns(int)
	% effects  ...
	end _gcd

end _adv


% _Byteptr

_byteptr = cluster is create, reset, next, peek, deposit, adjust, empty,
	   get_buf, get_index, get_count, equal, similar, copy, _gcd

% Operations

create = proc(BV: _bytevec, I, CNT: int) returns(_byteptr)
	 signals(bounds, negative_size)
	% start index i, cnt chars
	end create

reset = proc(BP: _byteptr, BV: _bytevec, I, CNT: int)
	signals(bounds, negative_size)
	% start index i, cnt chars
	end reset

next = proc(BP: _byteptr) returns(char) signals(bounds)
	% read char and increment index
	end next

peek = proc(BP: _byteptr) returns(char) signals(bounds)
	% peek at next char
	end peek

deposit = proc(BP: _byteptr, C: char) signals(bounds)
	% write char and increment index
	end deposit

adjust = proc(BP: _byteptr, DISP: int) signals(bounds)
	% move disp chars relative to current position
	end adjust

empty = proc(BP: _byteptr) returns(bool)
	% test if next read/write will fail
	end empty

get_buf = proc(BP: _byteptr) returns(_bytevec)
	end get_buf

get_index = proc(BP: _byteptr) returns(int)
	% get next index to be read/written
	end get_index

get_count = proc(BP: _byteptr) returns(int)
	% get count of remaining chars to read/write
	end get_count

equal = proc(BP1, BP2: _byteptr) returns(bool)
	% object identity
	end equal

similar = proc(BP1, BP2: _byteptr) returns(bool)
	% same buf, index, and count
	end similar

copy = proc(BP: _byteptr) returns(_byteptr)
	% really copy1
	end copy

_gcd = proc(BP: _byteptr, TAB: gcd_tab) returns(int)
	end _gcd

end _byteptr


% _Byteptr8

_byteptr8 = cluster is create, reset, next, peek, deposit, adjust, empty,
	    get_buf, get_index, get_count, equal, similar, copy, _gcd

% Operations

create = proc(WV: _wordvec, I, CNT: int) returns(_byteptr8)
	 signals(bounds, negative_size)
	% start index i, cnt bytes
	end create

reset = proc(BP: _byteptr8, WV: _wordvec, I, CNT: int)
	signals(bounds, negative_size)
	% start index i, cnt bytes
	end reset

next = proc(BP: _byteptr8) returns(int) signals(bounds)
	% read byte and increment index
	end next

peek = proc(BP: _byteptr8) returns(int) signals(bounds)
	% peek at next byte
	end peek

deposit = proc(BP: _byteptr8, I: int) signals(bounds)
	% write byte and increment index
	end deposit

adjust = proc(BP: _byteptr8, DISP: int) signals(bounds)
	% move disp bytes relative to current position
	end adjust

empty = proc(BP: _byteptr8) returns(bool)
	% test if next read/write will fail
	end empty

get_buf = proc(BP: _byteptr8) returns(_wordvec)
	end get_buf

get_index = proc(BP: _byteptr8) returns(int)
	% get next index to be read/written
	end get_index

get_count = proc(BP: _byteptr8) returns(int)
	% get count of remaining bytes to read/write
	end get_count

equal = proc(BP1, BP2: _byteptr8) returns(bool)
	% object identity
	end equal

similar = proc(BP1, BP2: _byteptr8) returns(bool)
	% same buf, index, and count
	end similar

copy = proc(BP: _byteptr8) returns(_byteptr8)
	% really copy1
	end copy

_gcd = proc(BP: _byteptr8, tab: gcd_tab) returns(int)
	end _gcd

end _byteptr8


% _Bytevec

_bytevec = cluster is create, fetch, store, size, equal, indexc, indexv,
	   nc_indexv, comp, copy, move_lr, move_rl, move_v, move_b, build, _gcd

% Operatinos

create = proc(SZ: int) returns(_bytevec) signals(toobig)
	% number of chars (all initially '\000')
	end create

copy = proc(V: _bytevec) returns(_bytevec)
	% really copies
	end copy

fetch = proc(V: _bytevec, I: int) returns(char) signals(bounds)
	% start index is 1
	end fetch

store = proc(V: _bytevec, I: int, C: char) signals(bounds)
	% start index is 1
	end store

size = proc(V: _bytevec) returns(int)
	% number of chars
	end size

equal = proc(V1, V2: _bytevec) returns(bool)
	% object identity
	end equal

indexc = proc(C: char, V: _bytevec, FIRST: int) returns(int) signals(bounds)
	% MIN{i | i >= first & v[i] = c}
	% returns 0 if no match
	end indexc

indexv = proc(V1, V2: _bytevec, FIRST: int) returns(int) signals(bounds)
	% MIN{i | i >= first & v1[j] = v2[i - 1 + j] for all j}
	% returns 0 if no match
	end indexv

nc_indexv = proc(V1, V2: _bytevec, FIRST: int) returns(int) signals(bounds)
	% same as indexv but case is ignored
	end nc_indexv

comp = proc(V1, V2: _bytevec) returns(int)
	% returns -1 if v1 < v2
	%	   0 if v1 = v2		(char comparison, not object identity)
	%	  +1 if v1 > v2
	end comp

move_lr = proc(V1: _bytevec, S1: int, V2: _bytevec, S2: int, LEN: int)
	  signals(negative_size, bounds)
	% moves len chars from v1 starting at s1 to v2 starting at s2
	% first char is moved first
	% 1 <= s1 <= size(v1) + 1
	% 1 <= s2 <= size(v2) + 1
	% len >= 0
	end move_lr

move_rl = proc(V1: _bytevec, S1: int, V2: _bytevec, S2: int, LEN: int)
	  signals(negative_size, bounds)
	% moves len chars from v1 starting at s1 to v2 starting at s2
	% last char is moved first
	% 1 <= s1 <= size(v1) + 1
	% 1 <= s2 <= size(v2) + 1
	% len >= 0
	end move_rl

move_v = proc(V1: _vec[char], S1: int, V2: _bytevec, S2: int, LEN: int)
	 signals(negative_size, bounds)
	% moves len chars from v1 starting at s1 to v2 starting at s2
	% 1 <= s1 <= size(v1) + 1
	% 1 <= s2 <= size(v2) + 1
	% len >= 0
	end move_v

move_b = proc(V1: _bytevec, S1: int, V2: _vec[char], S2: int, LEN: int)
	 signals(negative_size, bounds)
	% moves len chars from v1 starting at s1 to v2 starting at s2
	% 1 <= s1 <= size(v1) + 1
	% 1 <= s2 <= size(v2) + 1
	% len >= 0
	end move_b

build = proc(V: _bytevec) returns(_bytevec)
	% builds new _bytevec from ASCIZ string, converts to lower case
	end build

_gcd = proc(V: _bytevec, TAB: gcd_tab) returns(int)
	end _gcd

end _bytevec


% _Chan

_chan = cluster is open, _open, pipe, socket, socketpair, primary_input,
	primary_output, error_output, nul, control, lock, sync, reset, close,
	abort, is_tty, getc, geti, getb, getbv, getw, getwv, get, recv,
	recvfrom, pending, pend_count, buf_count, putc, puti, puts, putb, putw,
	put, send, sendto, read_wait, write_wait, wait, listen, accept, bind,
	connect, shutdown, peername, sockname, getsockopt, setsockopt,
	get_name, get_date, get_speeds, get_block_size, stat, get_access,
	set_access, get_length, set_length, equal, similar, copy,
	get_read_channel, get_write_channel, get_def_mode, set_def_mode,
	_save_tty, _restore_tty, close_all

% Operations

open = proc(FN: file_name, MODE: string, FMODE: int) returns(_chan)
       signals(not_possible(string))
	% modes are "read"		read existing file
	%	    "write"		overwrite existing file/write new file
	%	    "append"		append to existing file/write new file
	%	    "random"		overwrite & read existing/new file
	%	    "modify"		read & write existing/new file
	% fmode is protection mode for new files
	% terminal _chan is always read/write for both normal and image I/O
	end open

_open = proc(FN: file_name, FLAGS, FMODE: int) returns(_chan)
	signals(not_possible(string))
	% flags are as in Unix open system call
	% fmode is protection mode for new files
	% terminal _chan is always read/write for both normal and image I/O
	end _open

pipe = proc() returns(_chan, _chan) signals(not_possible(string))
	% returns read _chan, write _chan
	end pipe

socket = proc(FAMILY, SOCKTYPE, PROTOCOL: int) returns(_chan)
	 signals(not_possible(string))
	% create a socket
	end socket

socketpair = proc(DOMAIN, SOCKTYPE, PROTOCOL: int) returns(_chan, _chan)
	     signals(not_possible(string))
	% create a pair of sockets
	end socketpair

primary_input = proc() returns(_chan)
	% primary input
	% invoking this when both primary input and output are the terminal
	% will cause a tty mode change.  from that point on, the program cannot
	% be run in the background.
	end primary_input

primary_output = proc() returns(_chan)
	% primary output
	% invoking this when both primary input and output are the terminal
	% will cause a tty mode change.  from that point on, the program cannot
	% be run in the background.
	end primary_output

error_output = proc() returns(_chan)
	% error output
	end error_output

nul = proc() returns(_chan)
	% infinite source/sink
	end nul

control = proc(CH: _chan, CMD, ARG: int) returns(int)
	  signals(not_possible(string))
	% cmd and arg are as in fcntl Unix system call
	end control

lock = proc(CH: _chan, OP: int) signals(not_possible(string))
	% op is as in flock Unix system call
	end lock

sync = proc(CH: _chan) signals(not_possible(string))
	% flush in-core buffers to disk
	end sync

reset = proc(CH: _chan) signals(not_possible(string))
	% sets access pointer to 0
	end reset

close = proc(CH: _chan) signals(permanent, not_possible(string))
	% close _chan
	end close

abort = proc(CH: _chan) signals(permanent, not_possible(string))
	% close _chan, discard contents and file if new
	end abort

is_tty = proc(CH: _chan) returns(bool)
	% terminal
	end is_tty

getc = proc(CH: _chan, IMAGE: bool) returns(char)
       signals(end_of_file, not_possible(string))
	% read one byte as a char (image only applies to terminals)
	% ^M cannot be read from a terminal, even in image mode
	end getc

geti = proc(CH: _chan, IMAGE: bool) returns(int)
       signals(overflow, end_of_file, not_possible(string))
	% read one byte as an int (image only applies to terminals)
	% ^M cannot be read from a terminal, even in image mode
	end geti

getb = proc(CH: _chan, BV: _bytevec) returns(int)
       signals(bounds, end_of_file, not_possible(string))
	% read bytes into bv, return count read
	end getb

getbv = proc(CH: _chan, BV: _bytevec, STRT, N: int) returns(int)
	signals(bounds, end_of_file, not_possible(string))
	% read n bytes into bv starting at index strt
	% return count of bytes actually read
	end getbv

getw = proc(CH: _chan, WV: _wordvec) returns(int)
       signals(bounds, end_of_file, not_possible(string))
	% read bytes into wv, return count read
	end getw

getwv = proc(CH: _chan, WV: _wordvec, STRT, N: int) returns(int)
	signals(bounds, end_of_file, not_possible(string))
	% read n bytes into wv starting at byte index strt
	% return count of bytes actually read
	end getwv

get = proc(CH: _chan, BV: _bytevec, LOW, HIGH: int, TERM: string, IMAGE: bool)
      returns(string, int, int)
      signals(bounds, end_of_file, not_possible(string))
	% do stream-like gets, using bvec for buffer
	% low-high gives current non-empty portion of buffer
	% returns new low-high
	end get

recv = proc(CH: _chan, BUF: _wordvec, FLAGS: int) returns(int)
       signals(not_possible(string))
	% receive a message in buf
	% flags are as in recv Unix system call
	% returns length of message read
	end recv

recvfrom = proc(CH: _chan, BUF: _wordvec, FLAGS: int, ADDR: _wordvec)
	   returns(int, int) signals(not_possible(string))
	% receive a message in buf
	% flags are as in recvfrom Unix system call
	% sender address is stored in addr
	% returns length of message read and length of addr stored
	end recvfrom

pending = proc(CH: _chan, IMAGE: bool) returns(bool)
	  signals(not_possible(string))
	% tests if input is pending
	end pending

pend_count = proc(CH: _chan) returns(int) signals(not_possible(string))
	% returns number of bytes pending
	end pend_count

buf_count = proc(CH: _chan) returns(int) signals(not_possible(string))
	% returns number of buffered bytes
	end buf_count

putc = proc(CH: _chan, C: char, IMAGE: bool) signals(not_possible(string))
	% write char as one byte
	end putc

puti = proc(CH: _chan, I: int, IMAGE: bool) signals(not_possible(string))
	% write int as one byte
	end puti

puts = proc(CH: _chan, S: string, IMAGE: bool) signals(not_possible(string))
	% write chars as bytes
	end puts

putb = proc(CH: _chan, BV: _bytevec, STRT, STOP: int, IMAGE: bool)
       signals(bounds, not_possible(string))
	% write chars from bv starting at strt to stop
	end putb

putw = proc(CH: _chan, WV: _wordvec, STRT, STOP: int, IMAGE: bool)
       signals(bounds, not_possible(string))
	% write bytes from wv starting at byte index strt to stop
	end putw

put = proc(CH: _chan, BVEC: _bytevec, LOW, HIGH: int, S: string, IMAGE: bool)
      returns(int, int) signals(bounds, not_possible(string))
	% do stream-like puts using bvec as buffer
	% low-high gives current non-empty portion of buffer
	% returns new low-high
	end put

send = proc(CH: _chan, BUF: _wordvec, LEN, FLAGS: int)
       signals(bounds, not_possible(string))
	% send the message in the first len bytes of buf
	% flags are as in send Unix system call
	% signals bounds if len < 0
	end send

sendto = proc(CH: _chan, BUF: _wordvec, LEN, F: int, A: _wordvec, ALEN: int)
	 signals(bounds, not_possible(string))
	% send the message in the first len bytes of buf to the address
	%  in the first alen bytes of addr
	% flags are as in sendto Unix system call
	% signals bounds if len < 0 or alen < 0
	end sendto

read_wait = proc(CH: _chan, MILLIS: int) returns(bool)
	    signals(not_possible(string))
	% wait until either ch has data or an exception available, or
	%  until millis milliseconds have elapsed
	% returns true if ch has data or an exception available
	end read_wait

write_wait = proc(CH: _chan, MILLIS: int) returns(bool)
	     signals(not_possible(string))
	% wait until either ch can take data or an exception is available, or
	%  until millis milliseconds have elapsed
	% returns true if ch can take data or an exception is available
	end write_wait

wait = proc(READS, WRITES: sequence[_chan], MILLIS: int) returns(int)
       signals(not_possible(string))
	% wait until one or more of reads have data or an exception available
	%  or writes can take data or have an exception available, or until
	%  millis milliseconds have elapsed
	% returns number of ready _chans
	end wait

listen = proc(CH: _chan, BACKLOG: int) signals(not_possible(string))
	% allow socket connections
	% backlog is as in listen Unix system call
	end listen

accept = proc(CH: _chan, ADDR: _wordvec) returns(_chan, int)
	 signals(not_possible(string))
	% accept a connection on the socket
	% the address of the connector is stored in addr
	% returns a new socket and the length of the address stored in addr
	end accept

bind = proc(CH: _chan, NAME: _wordvec, LEN: int)
       signals(bounds, not_possible(string))
	% bind a name to a socket
	% len is the length of the name in name
	end bind

connect = proc(CH: _chan, ADDR: _wordvec, LEN: int)
	  signals(bounds, not_possible(string))
	% connect an addres to the socket
	% len is the length of the addres in addr
	end connect

shutdown = proc(CH: _chan, HOW: int) signals(not_possible(string))
	% shut down part of a connection
	% how is as in shutdown Unix system call
	end shutdown

peername = proc(CH: _chan, ADDR: _wordvec) returns(int)
	   signals(not_possible(string))
	% store name of connected peer in addr
	% returns length of stored name
	end peername

sockname = proc(CH: _chan, ADDR: _wordvec) returns(int)
	   signals(not_possible(string))
	% store name of socket in addr
	% returns length of stored name
	end sockname

getsockopt = proc(CH: _chan, LEVEL, OPTION: int) returns(int)
	     signals(not_possible(string))
	% get socket option
	% level and option are as in getsockopt Unix system call
	% returns value (assumed to be an int)
	end getsockopt

setsockopt = proc(CH: _chan, LEVEL, OPTION, VALUE: int)
	     signals(not_possible(string))
	% set socket option
	% level and option are as in setsockopt Unix system call
	end setsockopt

get_name = proc(CH: _chan) returns(file_name) signals(not_possible(string))
	% exact file name
	end get_name

get_date = proc(CH: _chan) returns(date) signals(not_possible(string))
	% get date last written
	end get_date

set_date = proc(CH: _chan, D: date) signals(not_possible(string))
	% set date last written
	end set_date

get_speeds = proc(CH: _chan) returns(int, int) signals(not_possible(string))
	% get input and output line speeds if terminal
	end get_speeds

get_block_size = proc(CH: _chan) returns(int) signals(not_possible(string))
	% return optimal block size
	end get_block_size

stat = proc(CH: _chan, BUF: _wordvec) signals(bounds, not_possible(string))
	% do fstat Unix system call storing result in buf
	end stat

get_access = proc(CH: _chan) returns(int) signals(not_possible(string))
	% get access pointer (starts at 0)
	end get_access

set_access = proc(CH: _chan, POS: int) signals(not_possible(string))
	% set access pointer to byte pos (starts at 0)
	end set_access

get_length = proc(CH: _chan) returns(int) signals(not_possible(string))
	% get length in bytes
	end get_length

set_length = proc(CH: _chan, LEN: int) signals(not_possible(string))
	% set length in bytes
	end set_length

equal = proc(CH1, CH2: _chan) returns(bool)
	% object identity
	end equal

similar = proc(CH1, CH2: _chan) returns(bool)
	% object identity
	end similar

copy = proc(CH: _chan) returns(_chan)
	% does not copy
	end copy

get_read_channel = proc(CH: _chan, IMAGE: bool) returns(int)
		   signals(not_possible(string))
	% get read jfn
	end get_read_channel

get_write_channel = proc(CH: _chan, IMAGE: bool) returns(int)
		    signals(not_possible(string))
	% get write jfn
	end get_write_channel

get_def_mode = proc() returns(int)
	% default protection for new files
	end get_def_mode

set_def_mode = proc(MODE: int)
	% default protection for new files
	end set_def_mode

_save_tty = proc()
	% save current tty mode and restore normal mode
	end _save_tty

_restore_tty = proc()
	% restore saved tty mode
	end _restore_tty

close_all = proc()
	% close all open file descriptors
	end close_all

end _chan


% _Event

_event = cluster is defer, undefer

% Operations

defer = proc()
	% defer quit interrupt, stop array-trimming during garbage collection
	% must be paired with undefer call!
	end defer

undefer = proc()
	% allow quit interrupt, allow array-trimming during garbage collection
	% must be paired with defer call!
	end undefer

end _event

% _Gc

_gc = cluster is count, control, gc

% Operations

count = proc() returns(int)
	% return number of garbage collections so far
	end count

control = proc(ADD, MASK: int) signals(illegal)
	% after a GC, new GC point is (heap_size+ADD)&~MASK
	% ADD > 0, MASK lssu ADD
	% default: ADD = 2**20-2**17, MASK = 2**19-1
	end control

gc = proc()
	% force a garbage collect
	end gc

end _gc


% _Job

_job = cluster is cons, cons1, cone, cone1, exec, exece, fork, run, sig,
       sig_group, dead, wait, wait0, any_wait, any_wait0, self, get_pid,
       get_group, set_group, control_tty, get_group_mode, set_group_mode,
       equal, similar, copy

% Operations

cons = proc(FN: file_name, JCL: string, PRI, PRO, ERO: _chan,
	    SAVE: sequence[_chan])
       returns(_job) signals(not_possible(string))
	% fork process to execute FN with given JCL
	% If JCL does not contain nuls, spaces and tabs will be converted
	% PRI, PRO, and ERO become primary input, output, and error output
	% SAVE is a list of _chans that should not be closed in the new process
	end cons

cons1 = proc(FN: file_name, JCL: string, SAVE: sequence[_chan]) returns(_job)
	signals(not_possible(string))
	% fork process to execute FN with given JCL
	% If JCL does not contain nuls, spaces and tabs will be converted
	% current primary input, output, and error output are retained
	% SAVE is a list of _chans that should not be closed in the new process
	end cons1

cone = proc(FN: file_name, JCL: string, ENVS: array[string],
	    PRI, PRO, ERO: _chan, SAVE: sequence[_chan])
       returns(_job) signals(not_possible(string))
	% fork process to execute FN with given JCL
	% If JCL does not contain nuls, spaces and tabs will be converted
	% PRI, PRO, and ERO become primary input, output, and error output
	% envs is the environment for the new program.
	% SAVE is a list of _chans that should not be closed in the new process
	end cone

cone1 = proc(FN: file_name, JCL: string, ENVS: array[string],
	     SAVE: sequence[_chan])
	returns(_job) signals(not_possible(string))
	% fork process to execute FN with given JCL
	% If JCL does not contain nuls, spaces and tabs will be converted
	% current primary input, output, and error output are retained
	% envs is the environment for the new program.
	% SAVE is a list of _chans that should not be closed in the new process
	end cone1

exec = proc(FN: file_name, JCL: string, SAVE: sequence[_chan])
       signals(not_possible(string))
	% execute FN with given JCL in current process
	% If JCL does not contain nuls, spaces and tabs will be converted
	% current primary input, output, and error output are retained
	% SAVE is a list of _chans that should not be closed
	end exec

exece = proc(FN: file_name, JCL: string, ENVS: array[string],
	     SAVE: sequence[_chan])
	signals(not_possible(string))
	% execute FN with given JCL in current process
	% If JCL does not contain nuls, spaces and tabs will be converted
	% current primary input, output, and error output are retained
	% envs is the environment for the new program.
	% SAVE is a list of _chans that should not be closed
	end exece

fork = proc() returns(_job, bool) signals(not_possible(string))
	% fork copy of current process
	% returns (child, false) in parent process
	% returns (parent, true) in child process
	end fork

run = proc(J: _job) returns(int) signals(not_possible(string))
	% wait for termination of given process
	% restart process whenever it stops
	% discard stop or termination of any intervening process
	% return termination status
	end run

sig = proc(J: _job, SIGNUM: int) signals(not_possible(string))
	% send signal to given process
	end sig

sig_group = proc(J: _job, SIGNUM: int) signals(not_possible(string))
	% send signal to all processes in given process's process group
	end sig_group

wait = proc(J: _job) returns(int) signals(not_possible(string))
	% wait for stop or termination of given process
	% discard termination of any intervening process
	% return stop or termination status
	end wait

wait0 = proc(J: _job) returns(int) signals(not_possible(string))
	% test for stop or termination of given process without hanging
	% discard termination of any intervening process
	% return stop or termination status if any, else -1
	end wait0

any_wait = proc() returns(_job, int) signals(not_possible(string))
	% wait for stop or termination of next process
	% return stop or termination status
	end any_wait

any_wait0 = proc() returns(_job, int) signals(not_possible(string))
	% test for stop or termination of next process without hanging
	% return _job (0 if none terminated) and stop or termination status
	end any_wait0

self = proc() returns(_job)
	% return current process
	end self

get_pid = proc(J: _job) returns(int)
	% get process id
	end get_pid

get_group = proc(J: _job) returns(int) signals(not_possible(string))
	% get process group
	end get_group

set_group = proc(J: _job, G: int) signals(not_possible(string))
	% set process group
	end set_group

control_tty = proc(J: _job) signals(not_possible(string))
	% make this process' group the controlling terminals' group
	end control_tty

get_group_mode = proc() returns(int)
	% get process group mode
	end get_group_mode

set_group_mode = proc(MODE: int)
	% set process group mode
	% 0 => do nothing
	% ~= 0 => cons and cons1 will set process group to be process id
	% < 0 => cons and cons1 will also change control of file desc 0
	%	 to new process group
	end set_group_mode

dead = proc() returns(int)
	% return count of processes that have died
	end dead

equal = proc(J1, J2: _job) returns(bool)
	end equal

similar = proc(J1, J2: _job) returns(bool)
	end similar

copy = proc(J: _job) returns(_job)
	end copy

end _job


% _Obj

_obj = cluster is is_ref, is_int_, is_real_, is_uninit, is_proc_, is_iter_,
       is_routine, is_vec, is_bvec, is_cell, is_anyize, is_adv,
       is_entry_vector, get_addr, get_left, get_right, lt, le, ge, gt, equal,
       similar, copy, print

% Operations

is_ref = proc(X: _obj) returns(bool)
	end is_ref

is_int_ = proc(X: _obj) returns(bool)
	end is_int_

is_real_ = proc(X: _obj) returns(bool)
	end is_real_

is_uninit = proc(X: _obj) returns(bool)
	end is_uninit

is_proc_ = proc(X: _obj) returns(bool)
	end is_proc_

is_iter_ = proc(X: _obj) returns(bool)
	end is_iter_

is_routine = proc(X: _obj) returns(bool)
	end is_routine

is_vec = proc(X: _obj) returns(bool)
	end is_vec

is_bvec = proc(X: _obj) returns(bool)
	end is_bvec

is_cell = proc(X: _obj) returns(bool)
	end is_cell

is_anyize = proc(X: _obj) returns(bool)
	end is_anyize

is_adv = proc(X: _obj) returns(bool)
	end is_adv

is_entry_vector = proc(X: _obj) returns(bool)
	end is_entry_vector

get_addr = proc(X: _obj) returns(int)
	end get_addr

get_left = proc(X: _obj) returns(int)
	end get_left

get_right = proc(X: _obj) returns(int)
	end get_right

lt = proc(X, Y: _obj) returns(bool)
	end lt

le = proc(X, Y: _obj) returns(bool)
	end le

ge = proc(X, Y: _obj) returns(bool)
	end ge

gt = proc(X, Y: _obj) returns(bool)
	end gt

equal = proc(X, Y: _obj) returns(bool)
	end equal

similar = proc(X, Y: _obj) returns(bool)
	end similar

copy = proc(X: _obj) returns(_obj)
	end copy

print = proc(X: _obj, ps: pstream)
	end print

end _obj


% _Resolve

_resolve = cluster is n2a

n2a = proc(NAME: string) returns(int, int)
      signals(not_found, no_server, timeout)
	end n2a

end _resolve


% _Tagcell

_tagcell = cluster[T: type] is create, get, set, set_value, copy, equal, _gcd

% Operations

create = proc(TG: int, VAL: T) returns(_tagcell[T]) signals(toobig)
	% create tagged cell
	end create

copy = proc(TT: _tagcell[T]) returns(_tagcell[T])
	% actually copy1
	end copy

get = proc(TT: _tagcell[T]) returns(int, T)
	% return tag and value
	end get

set = proc(TT: _tagcell[T], TG: int, VAL: T) signals(toobig)
	% set tag and value
	end set

set_value = proc(TT: _tagcell[T], VAL: T)
	% set value
	end set_value

equal = proc(T1, T2: _tagcell[T]) returns(bool)
	% object identity
	end equal

_gcd = proc(TT: _tagcell[T], TAB: gcd_tab) returns(int)
       where t has _gcd: proctype(t, gcd_tab) returns(int)
	end _gcd

end _tagcell


% _Vec

_vec = cluster[T: type] is create, fetch, store, size, equal, copy, move_lr,
       move_rl, _gcd

% Overview
%
% An object of the type _vec[T] (hereafter called a vector) is a fixed-size,
% mutable collection of objects of the type T.  For each vector V, the size of
% V is the number of the elements of V.  Each element of a vector V is indexed
% by an integer starting from 1.

% Operations

create = proc(SZ: int) returns(_vec[T]) signals(toobig)
	% effects  Signals `toobig' if SZ is greater than an implementation-
	%	   defined limit; otherwise returns a new vector V such that
	%	   the size of V is SZ and all the elements are the same
	%	   value, which is guaranteed to be zero if T is `int'.
	end create

fetch = proc(VEC: _vec[T], INDEX: int) returns(T) signals(bounds)
	% effects  Signals `bounds' if INDEX is less than 1 or greater than the
	%	   size of VEC; otherwise returns the element at the position
	%	   INDEX in VEC.
	end fetch

store = proc(VEC: _vec[T], INDEX: int, VAL: T) signals(bounds)
	% modifies VEC
	% effects  Signals `bounds' if INDEX is less than 1 or greater than the
	%	   size of VEC; otherwise changes VEC so that the element at
	%	   the position INDEX becomes VAL.
	end store

size = proc(VEC: _vec[T]) returns(int)
	% effects  Returns the size (i.e., the number of elements) of VEC.
	end size

equal = proc(VEC1, VEC2: _vec[T]) returns(bool)
	% effects  Returns `true' if VEC1 and VEC2 are the same object;
	%	   otherwise returns `false'.
	end equal

copy = proc(VEC: _vec[T]) returns(_vec[T])
	% effects  Returns a new vector that has the same size and elements as
	%	   those of VEC.
	end copy

move_lr = proc(VEC1: _vec[T], I1: int, VEC2: _vec[T], I2: int, LEN: int)
	  signals(negative_size, bounds)
	% modifies VEC2
	% effects  Signals `negative_size' if LEN < 0; signals `bounds' if
	%	   I1 < 1 or I1 > size(VEC1) or I2 < 1 or I2 > size(VEC2);
	%	   otherwise changes VEC2 so that the element at the position
	%	   I2 + OFFSET becomes VEC1[I1 + OFFSET] for every OFFSET such
	%	   that 0 <= OFFSET < min{LEN, MAX1, MAX2}, where
	%	   MAX1 = size(VEC1) - I1 +1 and MAX2 = size(VEC2) - I2 + 1.
	%	   Copying is performed in the "left-to-right" order.
	end move_lr

move_rl = proc(VEC1: _vec[T], I1: int, VEC2: _vec[T], I2: int, LEN: int)
	  signals(negative_size, bounds)
	% modifies VEC2
	% effects  Same as the `move_lr' operation except that copying is
	%	   performed in the "right-to-left" order.
	end move_rl

_gcd = proc(V: _vec[T], TAB: gcd_tab) returns(int)
       where T has _gcd: proctype(T, gcd_tab) returns(int)
	% effects  ...
	end _gcd

end _vec


% _Wordvec

_wordvec = cluster is create, fetch, store, fetch2, store2, xstore, bfetch,
	   bstore, wfetch, wstore, lfetch, lstore, move_lr, move_rl, get_byte,
	   set_byte, word_size, byte_size, bytes_per_word, move_b2w, move_w2b,
	   size, equal, copy, _gcd

% Operations

create = proc(N: int) returns(_wordvec) signals(toobig)
	% number of words
	end create

fetch = proc(WV: _wordvec, IDX: int) returns(int) signals(bounds, overflow)
	% start index is 1
	end fetch

store = proc(WV: _wordvec, IDX, I: int) signals(bounds)
	% start index is 1
	end store

fetch2 = proc(WV: _wordvec, IDX: int) returns(int, int) signals(bounds)
	% returns left and right halves sign extended
	end fetch2

store2 = proc(WV: _wordvec, IDX, L, R: int) signals(bounds)
	% stores left and right halves
	end store2

xstore = proc(WV: _wordvec, IDX, B, L: int) signals(bounds)
	% stores high byte and low 24-bit values
	end xstore

bfetch = proc(WV: _wordvec, IDX: int) returns(int) signals(bounds)
	% start byte index is 1
	end bfetch

bstore = proc(WV: _wordvec, IDX, I: int) signals(bounds)
	% start byte index is 1
	end bstore

wfetch = proc(WV: _wordvec, IDX: int) returns(int) signals(bounds)
	% start byte index is 1
	end wfetch

wstore = proc(WV: _wordvec, IDX, I: int) signals(bounds)
	% start byte index is 1
	end wstore

lfetch = proc(WV: _wordvec, IDX: int) returns(int) signals(bounds, overflow)
	% start byte index is 1
	end lfetch

lstore = proc(WV: _wordvec, IDX, I: int) signals(bounds)
	% start byte index is 1
	end lstore

move_lr = proc(WV1: _wordvec, S1: int, WV2: _wordvec, S2: int, LEN: int)
	  signals(negative_size, bounds)
	% moves len words from WV1 starting at s1 to WV2 starting at s2
	% first word is moved first
	% 1 <= s1 <= size(WV1) + 1
	% 1 <= s2 <= size(WV2) + 1
	% len >= 0
	end move_lr

move_rl = proc(WV1: _wordvec, S1: int, WV2: _wordvec, S2: int, LEN: int)
	  signals(negative_size, bounds)
	% moves len words from WV1 starting at s1 to WV2 starting at s2
	% last word is moved first
	% 1 <= s1 <= size(WV1) + 1
	% 1 <= s2 <= size(WV2) + 1
	% len >= 0
	end move_rl

get_byte = proc(WV: _wordvec, WRD, BIT, LEN: int) returns(int)
	   signals(bounds, overflow, illegal_size)
	% get len bits ending at bit bit at index wrd
	% most significant bit is 1
	end get_byte

set_byte = proc(WV: _wordvec, BYTE, WRD, BIT, LEN: int)
	   signals(bounds, illegal_size)
	% set len bits ending at bit bit at index wrd
	% most significant bit is 1
	end set_byte

word_size = proc() returns(int)
	% returns 32 on VAX
	end word_size

byte_size = proc() returns(int)
	% returns 8 on VAX
	end byte_size

bytes_per_word = proc() returns(int)
	% returns 4 on VAX
	end bytes_per_word

move_b2w = proc(B: _bytevec, BPOS: int, W: _wordvec, WPOS: int, N: int)
	   signals(bounds)
	% moves n bytes from b starting at bpos to w starting at wpos
	end move_b2w

move_w2b = proc(W: _wordvec, WPOS: int, B: _bytevec, BPOS: int, N: int)
	   signals(bounds)
	% moves n bytes from w starting at wpos to b starting at bpos
	end move_w2b

size = proc(W: _wordvec) returns(int)
	% number of words
	end size

equal = proc(WV1, WV2: _wordvec) returns(bool)
	% object identity
	end equal

copy = proc(W: _wordvec) returns(_wordvec)
	% really copies
	end copy

_gcd = proc(W: _wordvec, TAB: gcd_tab) returns(int)
	end _gcd

end _wordvec


% Gcd_store

gcd_store = cluster is init, store_id, store_lit, store_ref, store_hdr,
	    store_array_desc, store_string, done

init = proc(W: _wordvec, C:_chan)
	end init

store_id = proc()
	end store_id

store_lit = proc(I1: int)
	end store_lit

store_ref = proc(I1: int)
	end store_ref

store_hdr = proc(HID, SIZE: int)
	end store_hdr

store_array_desc = proc(A: _adv[_obj])
	end store_array_desc

store_string = proc(S: _bytevec)
	end store_string

done = proc()
	end done

need = proc(I:int) returns (int)
	end need

end gcd_store


% Gcd_tab

gcd_tab = cluster is grind, grindc, grindp

% Operations

grind = proc[T: type](X: T, FN: file_name) returns(file_name)
	signals (not_possible(string))
	where T has _gcd: proctype(T, gcd_tab) returns(int)
	end grind

grindc = proc[T: type](X: T, CH: _chan, POS: int)
	 signals(not_possible(string))
	 where T has _gcd: proctype(T, gcd_tab) returns(int)
	end grindc

grindp = proc[T: type](X: T, CH: _chan) signals(not_possible(string))
	 where T has _gcd: proctype(T, gcd_tab) returns(int)
	end grindp

end gcd_tab


% Procedural/control abstractions


_anyize = proc[T: type](X: T) returns(any)
	end _anyize

_change_signal = proc(SIG: int, LABEL: int) returns(int)
		 signals(not_possible(string))
	% change action for given signal(interrupt), return old action
	end _change_signal

_cleanup_ = proc()
	end _cleanup_

_cvt = proc[FROM, TO: type](X: FROM) returns(TO)
	% fake compiler out as to type of object (use with great care!)
	end _cvt

_dtable_size = proc() returns(int)
	% returns maximum number of open files
	end _dtable_size

_environ = proc(NAME: string) returns(string) signals(not_found)
	% return environment variable
	end _environ

_environs = iter() yields(string)
	% yield the environment
	end _environs

_ernum = proc(MSG: string) returns(int) signals(not_found)
	% return error code for given error string
	end _ernum

_erstr = proc(NUM: int) returns(string)
	% return error string for given error code
	end _erstr

_fixup_dir_part = proc(S: string) returns(string)
	end _fixup_dir_part

_fixup_file_name = proc(FN: file_name, DIROK: bool) returns(file_name)
	end _fixup_file_name

_free_space = proc() returns(int)
	% number of free (long) words left in heap
	end _free_space

_get_gid = proc() returns(int, int)
	% real group ID, effective group ID
	end _get_gid

_get_pid = proc() returns(int)
	% process ID
	end _get_pid

_get_priority = proc(WHICH, WHO: int) returns(int)
		signals(not_possible(string))
	% do getpriority Unix system call
	end _get_priority

_get_termcap = proc() returns(string) signals(not_found)
	end _get_termcap

_get_uid = proc() returns(int, int)
	% real user ID, effective user ID
	end _get_uid

_groups = proc() returns(sequence[int])
	% returns group access list
	end _groups

_heap_size = proc() returns(int)
	% current size of heap in (long) words
	end _heap_size

_home_dir = proc(USER: string) returns(string) signals(not_found)
	end _home_dir

_host_id = proc() returns(int, int)
	% returns upper and lower halves of host id
	end _host_id

_ioctl = proc[T: type](FDESC, REQH, REQL: int, ARGP: T, DISP: int)
	 signals(not_possible(string))
	% issue REQ ioctl on given file descriptor with given argument(s)
	% REQ is given in halves, high and low
	% if ARGP is not an int, DISP should generally be 4
	end _ioctl

_last_indexc = proc(C: char, S: string) returns(int)
	% Find the last occurence of c in the string s.
	end _last_indexc

_lstat = proc(FN: file_name, BUF: _wordvec)
	 signals(bounds, not_possible(string))
	% do lstat Unix system call placing result in buf
	end _lstat

_page_size = proc() returns(int)
	% returns system page size in bytes
	end _page_size

_pause = proc()
	% wait for signal (interrupt)
	end _pause

_read_link = proc(FN: file_name, BUF: _bytevec) returns(int)
	     signals (not_possible(string))
	% read value of symbolic link into buf
	% returns number of characters read
	end _read_link

_round_real = proc(DIGS: _bytevec, NUM: int) returns(int, int)
	% round DIGS to NUM digits
	% used by e_form, f_form, g_form
	end _round_real

_rusage = proc(WHO: int, BUF: _wordvec) signals(bounds, not_possible(string))
	% do getrusage Unix system call, placing result in buf
	end _rusage

_select = proc(CNT, READS, WRITES, EXCEPTS, MILLIS: int)
	  returns(int, int, int, int)
	  signals(not_possible(string))
	% do select Unix system call
	% returns (number found, read mask, write mask, except mask)
	end _select

_set_alarm = proc(SECS: int) returns(bool)
	% set alarm interrupt to go off in SECS seconds real time
	% (0 => no alarm)
	% cancels any previous alarm
	% returns true if alarm has gone off since last call
	end _set_alarm

_set_limit = proc(RESOURCE, VALUE: int) signals(not_possible(string))
	% set soft resource limit via setrlimit Unix system call.
	end _set_limit

_set_priority = proc(WHICH, WHO: int) returns(int)
		signals(not_possible(string))
	% do setpriority Unix system call
	end _set_priority

_signal_type = proc(SIG: int) returns(string) signals(illegal_signal)
	end _signal_type

_skip_fill = proc(B: _bytevec, MAX: int, C: _chan, TERM: char) returns(int)
	     signals(end_of_file)
	end _skip_fill

_split_name_part = proc(S: string) returns(string, string, string)
	% Split file name up into name, suffix, rest (Separator = '.')
	end _split_name_part

_stat = proc(FN: file_name, BUF: _wordvec)
	signals(bounds, not_possible(string))
	% do stat Unix system call placing result in buf
	end _stat

_termcap = proc(TERM, CAP: string, REPEAT, BAUD: int) returns(string)
	   signals(not_found)
	end _termcap

_termcap_decode = proc(TERM: string, I: int) returns(int, char)
		  signals(bounds)
	end _termcap_decode

_termcap_entry = proc(FS, TERM: string) returns(string) signals(not_found)
	% Return the termcap entry for fs in terminal term.
	end _termcap_entry

_umask = proc(MASK: int) returns(int) signals(not_possible(string))
	% effects  Invokes the `umask' system call with MASK.  Returns the
	%	   result of the call if the call was successful; otherwise
	%	   signals `not_possible'.
	end _umask

_unparse_real = proc(X: real) returns(bool, _bytevec, int)
	% effects  The first result is the sign of X: `true' if X is negative;
	%	   `false' otherwise.  The second result is the sequence of
	%	   decimal digits of the fraction part of X.  The third result
	%	   is the exponent of X.
	end _unparse_real

_working_dir = proc(S: string, UPDATE: bool) returns(string)
	       signals(bad_format)
	% effects  Signals `bad_format' if UPDATE is `true' and S is not a
	%	   valid as a file name; sets the working directory to S and
	%	   returns the working directory if UPDATE is `true' and S is
	%	   valid as a file name; otherwise returns the current working
	%	   directory.
	end _working_dir

inet_address = proc(SPEC: string) returns(int, int) signals(bad_format)
	end inet_address

istream2chan = proc(IST: istream) returns(_chan) signals(not_possible(string))
	% effects  Returns the `_chan' object derived from IST if possible;
	%	   otherwise signals `not_possible'.
	end istream2chan

stream2chan = proc(ST: stream) returns(_chan) signals(not_possible(string))
	% effects  Returns the `_chan' object derived from ST if possible;
	%	   otherwise signals `not_possible'.
	end stream2chan
